

#ifndef _OREGANO_TSE_MAC_H_
#define _OREGANO_TSE_MAC_H_

#include "compat/time.h"
#include "delay_correction.h"
#include "syn1588nic_reg.h"
#include "altera_pci_bridge.h"

#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/mii.h>
#include <linux/netdevice.h>
#include <linux/semaphore.h>
#include <linux/types.h>


struct oregano_mac_platform_data {
	struct device *pci_bridge; // nullable
	struct device *clock_m; // nullable
	struct i2c_adapter *i2c_adapter;

	u8 mac_address[ETH_ALEN];
};



#define MAC_BASE 0x00000000 ///< PCI NIC MAC base address
/// length of Ethernet VLAN tag in bytes
#define ETH_VLANTAG_LEN 4

#define NTP_PTP_OFFSET 0x83AA7E80

// I WANT TO KNOW MAXIMUM BHH PROCESSING TIME AND TX_QUEUE WAIT TIME
#define DEBUG_BHH_TX_QUEUE
#ifdef DEBUG_BHH_TX_QUEUE
#define DEBUG_BHH_MINIMUM_REPORT_DELTA 50000
#endif


#define BUFF_CNT 64 ///< number of ring buffer elements (HW/driver)
#if (BUFF_CNT <= 0 || BUFF_CNT * 2 > MAC_TOTAL_BD)
#error Illegal number of buffer descriptors!
#endif

/// desired MDC frequency [Hz] (period min 400ns according to IEEE 802.3 /
/// 22.2.2.11)
#define MDC_FRQ 500000 ///< Set MDC frequency
#ifndef MDC_FRQ
#define MDC_FRQ 5000000
#endif

#define MAC_FRQ 25000000 ///< oscillator frequency [Hz]
#if (MAC_FRQ / MDC_FRQ > 254 || MAC_FRQ / MDC_FRQ < 2)
#error MDC frequency out of dividor range! Adjust MAC_FRQ and/or MDC_FRQ!
#endif

/// DMA buffer size, which has to be at least a multiple of the PCI burst length
/// The DMA buffer size has to accomodate: ETH_FRAME_LEN + ETH_FCSLEN + RX
/// Timestamp + extra PCI burst
/// Just use 2k to be on the save side
// #define JUMBO_PACKET
#ifdef JUMBO_PACKET
#define DMA_BD_SIZE ((unsigned int)(1024 * 16))
#define ETH_MAX_JUMBO_FRAME_SIZE (DMA_BD_SIZE - ETH_VLANTAG_LEN - ETH_FCS_LEN)
#else
#define DMA_BD_SIZE ((unsigned int)(1024 * 2))
#endif

/**
 * \name General driver settings
 *@{
 */
#define PHY_ID_MASK 0x1F ///< mask for phy address
#define PHY_RN_MASK 0x1F ///< mask for PHY register number

/// buffer descriptor address structure
struct bd_addr {
	void __iomem *vaddr; ///< virtual address
	dma_addr_t addr; ///< bus address
};

/// structure for additional stats maintained by device driver
struct nic_extra_stats {
	unsigned long rx_ctrl_packets; ///< received control packets
	unsigned long rx_vlan_packets; ///< received VLAN tagged packets


	unsigned long hw_rx_packets;
	unsigned long hw_tx_packets;
	unsigned long hw_tx_error;
	unsigned long hw_vlan_packets;
	unsigned long hw_broadcast_packets;
	unsigned long hw_multicast_packets;

	unsigned long
		mcast_dropped; ///< multicast errors															// TODO: check
	unsigned long rx_swcrc_err; ///< error in SW CRC calculation
	unsigned long rx_late_coll; ///< receive path detected late collision
	unsigned long drv_rx_qfull; ///< driver receive queue full
	unsigned long drv_tx_qfull; ///< driver transmit queue full

	unsigned long
		tx_pkt_deferred; ///< MAC had to wait for carrier before sending packet

	unsigned long vlan_err; ///< receive packets with VLAN ID mismatch

	unsigned long ts_rx; ///< received timestamped packets
	unsigned long ts_tx; ///< transmitted timestamped packets
	unsigned long ts2_rx; ///< received timestamped packets
	unsigned long ts2_tx; ///< transmitted timestamped packets
	unsigned long ts_ptp_rx;
	unsigned long ts_ptp_tx;
};

struct clock_m_device;

enum OREGANO_MAC_IRQ {
	OREGANO_MAC_IRQ_LEGACY = 0,
	OREGANO_MAC_IRQ_AMOUNT,
};

/// NIC private data structure (MAC device)
struct nic_priv_data;
struct nic_priv_data *dev_get_mac(struct device *);


// depricated functions
u32 oregano_mac_readl(struct nic_priv_data *dev, u32 offset);
void oregano_mac_writel(struct nic_priv_data *dev, u32 val, u32 offset);

__must_check int nic_setReg(struct net_device *, unsigned long, u32, int);
int nic_getReg(struct net_device *, unsigned long, u32 *);


irqreturn_t nic_interrupt(int irq, void *dev_id);

#endif // _OREGANO_TSE_MAC_H_
