

#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/pci.h>
#include <linux/string.h>

#if defined(CONFIG_OF)
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/of_platform.h>
#endif

#include "compat/compat.h"
#include "syn1588nic_reg.h"
#include "oregano_spi.h"
#include "oregano_tse_mac.h"
#include "clock_m.h"
#include "si_time_pll.h"
#include "SyncDriver.h"
#include "version.h"

#define DRV_NAME "syn1588_drv"




#define SYN1588NIC_IRQ_MSI_MAC 0
#define SYN1588NIC_IRQ_MSI_CLOCK_M 0

#if (KERNEL_VERSION(3, 0, 0) <= LINUX_VERSION_CODE)
#define PDATA_SIZE pdata_size
#else
#define PDATA_SIZE data_size
#endif


#define PCI_VENDOR_ID_ALTERA 0x1172
#define PCI_DEVICE_ID_ALTERA_NIC_OLD_DEVICE_ID 0x1000
#define PCI_VENDOR_ID_MEINBERG 0x1360

#define PCI_DEVICE_ID_ALTERA_PCI_DEVICE_ID_ALTERA_NIC_OLD_DEVICE_ID 0x1000
#define PCI_DEVICE_ID_OREGANO_TEST_DEVICE 0x0a00
#define PCI_DEVICE_ID_MEINBERG_OREGANO_REV2_3_DEVICE 0x0a01
#define PCI_DEVICE_ID_MEINBERG_OREGANO_DUAL_NIC_REV1_1_DEVICE 0x0a06

struct syn1588_pci_nic_device {

	struct pci_dev *pci_dev; ///< PCI data structure
	struct syn1588_dev syn1588_dev;

	struct oregano_system_cabilities{
		bool has_oxco; // OCXO present
		bool is_sfp; //SFP type fiber network interface
		bool has_mac_serial; // MAC address and serial number device present
	} system_capabilities;

	//resources

	struct platform_device *pll_spi;
	struct platform_device *pci_bridge;


	struct si_time_95147 *si_time_pll;
};

struct syn1588_pci_data {
	const uint16_t pci_rev_max_supported;
	const uint16_t pci_rev_min_supported;
	const int num_macs;
};

static struct si_time_pll_platform_data rev_2_3_nic_pll_platform_data = {
	.clock_input_pll = 'A',
	.clock_input = 0,
	.clock_adjust_pll = 'C',
	.external_clock_input = 2,
	.sync_e_pll = 'D',
	.sync_e_clock_input = 3,
};

//Register information about your slave device:
struct spi_board_info spi_device_info[] = {
	{
		.modalias = "sit95147",
		.max_speed_hz = 10000000,
		.chip_select = 0,
		.mode = SPI_MODE_0,
		.platform_data = &rev_2_3_nic_pll_platform_data,
	}
};

static struct oregano_spi_platform_data  spi_platform_data = {
	.mode_bits = SPI_MODE_0 | SPI_CS_HIGH,
	.num_chipselect = 1,
	.devices = spi_device_info,
	.num_devices = ARRAY_SIZE(spi_device_info),
};



static int detect_type(struct pci_dev *pdev) {
	struct device *dev = &pdev->dev;
	void __iomem *io_base_bar0 = pcim_iomap_table(pdev)[0];
	struct syn1588_pci_nic_device *pci_data = pci_get_drvdata(pdev);

	if(!io_base_bar0) {
		dev_err(dev, "coould not iomap memory\n");
		return -ENOMEM;
	}
	pci_data->syn1588_dev.build_id = 999; // falback build_id
	if(pdev->vendor == PCI_VENDOR_ID_ALTERA) {
		u32 system_capabilities;
		pci_data->syn1588_dev.build_id = ioread32(io_base_bar0 + PCI_OFFSET_MAC + MAC_ETH_BUILD);
		if ( pci_data->syn1588_dev.build_id < 808 ) {
			dev_err(dev, "bitstream version (%d) is to old, please contact oregano support for an update\n", pci_data->syn1588_dev.build_id);
			return -ENODEV;
		}
		if(pci_data->syn1588_dev.build_id >= 2000) {
			pci_data->syn1588_dev.product_id = SYN1588_PRODUCT_ID_DUALNIC_REV1_0;
			dev_warn(dev, "old dual nic bitstream found, fallback to 1 interface\n");
		} else {
			u32 clock_id_h = ioread32(io_base_bar0 + PCI_OFFSET_CLOCK_M + CLOCKID_H);
			if((clock_id_h & ~0xFFu )  == 0x8CA5A100u) {
				pci_data->syn1588_dev.product_id = SYN1588_PRODUCT_ID_NIC_REV2_1;
			}else {
				pci_data->syn1588_dev.product_id = SYN1588_PRODUCT_ID_NIC_REV2_0;
			}
		}
		system_capabilities = ioread32(io_base_bar0 + PCI_OFFSET_SYSTEM_REGISTER_BLOCK + SYN1588_SYSTEM_REGISTER_BLOCK_CAPABILITY0);
		pr_debug("system_capabilities0: %08x", system_capabilities);
		if(test_bit(SYN1588_SYSTEM_REGISTER_BLOCK_CAPABILITY0_OXCO, (unsigned long *)&system_capabilities))
			pci_data->system_capabilities.has_oxco = true;
		if(test_bit(SYN1588_SYSTEM_REGISTER_BLOCK_CAPABILITY0_SFP, (unsigned long *)&system_capabilities))
			pci_data->system_capabilities.is_sfp = true;
		if(test_bit(SYN1588_SYSTEM_REGISTER_BLOCK_CAPABILITY0_MAC_SERIAL, (unsigned long *)&system_capabilities))
			pci_data->system_capabilities.has_mac_serial = true;
		return 0;
	}
	if(pdev->vendor == PCI_VENDOR_ID_MEINBERG) {
		char buf[16+1] = {};
		int i;
		pci_data->syn1588_dev.build_id = ioread32(io_base_bar0 + PCI_OFFSET_SYSTEM_REGISTER_BLOCK + SYN1588_SYSTEM_REGISTER_BLOCK_BUILD_ID);
		// if(pdev->device == syn1588_dev_pci_device_id(&syn1588nic_rev2_testing)) {
		// 	pci_data->syn1588_dev.product = &syn1588nic_rev2_testing;
		// } else
		if(pdev->device == PCI_DEVICE_ID_MEINBERG_OREGANO_REV2_3_DEVICE) {
			u32 board_capabilities = ioread32(io_base_bar0 + PCI_OFFSET_EEPROM + SYN1588_EEPROM_BOARD_CAPABILITIES);
			if((board_capabilities & SYN1588_EEPROM_BOARD_CAPABILITIES_FPGA_TYPE) == 0) {
				pci_data->syn1588_dev.product_id = SYN1588_PRODUCT_ID_NIC_REV2_3_A;
			} else {
				pci_data->syn1588_dev.product_id = SYN1588_PRODUCT_ID_NIC_REV2_3_B;
			}
		} else if(pdev->device == PCI_DEVICE_ID_MEINBERG_OREGANO_DUAL_NIC_REV1_1_DEVICE) {
			u32 system_config0 = ioread32(io_base_bar0 + PCI_OFFSET_SYSTEM_REGISTER_BLOCK + SYN1588_SYSTEM_REGISTER_BLOCK_CONFIG0);
			iowrite32(system_config0 | 0x1,io_base_bar0 + PCI_OFFSET_SYSTEM_REGISTER_BLOCK + SYN1588_SYSTEM_REGISTER_BLOCK_CONFIG0);
			pci_data->syn1588_dev.product_id = SYN1588_PRODUCT_ID_DUALNIC_REV1_1;
		} else {
			dev_err(dev, "unknown PCI card");
			return -ENODEV;
		}
		for(i=0; i < 4; i++) {
			u32 data = ioread32(io_base_bar0 + PCI_OFFSET_EEPROM + SYN1588_EEPROM_SERIAL_NR + i*4);
			buf[i*4 + 0] = (data >> 0)  & 0xff;
			buf[i*4 + 1] = (data >> 8)  & 0xff;
			buf[i*4 + 2] = (data >> 16) & 0xff;
			buf[i*4 + 3] = (data >> 24) & 0xff;
		}
		// trim leading space (' ') and NULL (0x00)
		for(i = strlen(buf); i >=0; i--) {
			if(buf[i] != ' ' && buf[i] != 0x00) {
				break;
			}
			buf[i] = 0x00;
		}
		strncpy(pci_data->syn1588_dev.serial_nr, buf, sizeof(pci_data->syn1588_dev.serial_nr)-1);

	}

	return 0;
}

#define DEFINE_PCI_RES_MEM_NAMED(_start, _size, _name)  \
	DEFINE_RES_MEM_NAMED((pdev->resource[0].start + (_start)), (_size), (_name))

static int syn1588_pci_probe(struct pci_dev *pdev, const struct pci_device_id *id) {
	struct device *dev = &pdev->dev;
	const struct syn1588_pci_data *syn1588_pci_data = (struct syn1588_pci_data *)id->driver_data;
	struct syn1588_pci_nic_device *pci_data = devm_kzalloc(dev, sizeof(struct syn1588_pci_nic_device), GFP_KERNEL);
	int err = 0;

	dev_info(dev, "load Oregano Syn1588 Nic pci driver version: " DRV_VERSION "\n");
	pci_set_drvdata(pdev, pci_data);

	if (pcim_enable_device(pdev)) {
		pr_err( "could not enable device\n");
		return -ENODEV;
	};

	err = dma_set_mask(dev, DMA_BIT_MASK(32));
	if (err) {
		dev_err(dev,  "can't set pci_dma_set_mast to 32bit\n");
		return -ENODEV;
	} else {
		err = dma_set_mask_and_coherent(dev, DMA_BIT_MASK(32));
		if (err) {
			dev_err(dev,  "could not set dma_mask\n");
			return -ENODEV;
		}
	}

	if(resource_size(&pdev->resource[0]) < PCI_SIZE_SYN1588_NIC) {
		dev_err(dev, "invalid bar size, to small\n");
		return -ENODEV;
	}

	pci_set_master(pdev); // enable busmastering


	//pci_request_region
	err = pcim_iomap_regions(pdev, 1 << 0, DRV_NAME);
	if(err) {
		dev_err(dev, "can not request region");
		return err;
	}

	err = detect_type(pdev);

	pcim_iounmap_regions(pdev, 1 << 0);

	if(
		err ||
		syn1588_pci_data->pci_rev_min_supported > pdev->revision ||
		syn1588_pci_data->pci_rev_max_supported < pdev->revision
	) {
		pr_err("unsupported nic, please contact orgegano support to get an update\n");
		pr_err("nic_procduct_id: %s, min_rev: %d, max_rev: %d, rev: %d\n",
			syn1588_dev_get_product_id(pci_data->syn1588_dev.product_id),
			syn1588_pci_data->pci_rev_min_supported,
			syn1588_pci_data->pci_rev_max_supported,
			pdev->revision
		);
		return -ENODEV;
	}
	dev_info(dev, "product name %s\n", syn1588_dev_get_product_id(pci_data->syn1588_dev.product_id));
	dev_info(dev, "build id %d\n", pci_data->syn1588_dev.build_id);

	if(pci_data->syn1588_dev.product_id == SYN1588_PRODUCT_ID_NIC_REV2_3_A || pci_data->syn1588_dev.product_id == SYN1588_PRODUCT_ID_NIC_REV2_3_B) {
		// new rev23 cards with new spi controller
		const struct resource syn1588_pci_spi_resources[] = {
			DEFINE_PCI_RES_MEM_NAMED(PCI_OFFSET_SPI_PLL, PCI_SIZE_SPI_PLL, "main"),
		};
		pci_data->pll_spi = platform_device_register_resndata(
			dev, "oregano-spi", PLATFORM_DEVID_AUTO, syn1588_pci_spi_resources, ARRAY_SIZE(syn1588_pci_spi_resources),
			&spi_platform_data, sizeof(spi_platform_data)
		);
		if(IS_ERR_OR_NULL(pci_data->pll_spi)) {
			dev_err(dev, "can not add pll spi\n");
			return PTR_ERR(pci_data->pll_spi);
		}
		pci_data->si_time_pll = sit_95147_get_device(&pci_data->pll_spi->dev);
	} else if(pci_data->syn1588_dev.product_id == SYN1588_PRODUCT_ID_DUALNIC_REV1_1) {
	} else if(id->vendor  == PCI_VENDOR_ID_ALTERA) {
		// old pci cards
		const struct resource syn1588_pci_bridge_resources[] = {
			DEFINE_PCI_RES_MEM_NAMED(PCI_OFFSET_PCI_BRIDGE, PCI_SIZE_PCI_BRIDGE, "main"),
		};
		pci_data->pci_bridge = platform_device_register_resndata(
			dev, "altera-pci-bridge", PLATFORM_DEVID_AUTO, syn1588_pci_bridge_resources, ARRAY_SIZE(syn1588_pci_bridge_resources),
			NULL, 0
		);
		if(IS_ERR_OR_NULL(pci_data->pci_bridge)) {
			dev_err(dev, "can not add pci_bridge\n");
			return PTR_ERR(pci_data->pci_bridge);
		}
	}

#if  KERNEL_VERSION(4, 8, 0) <= LINUX_VERSION_CODE
	//err = pci_alloc_irq_vectors(pdev, 9, 16, PCI_IRQ_MSI | PCI_IRQ_MSIX);
	err = pci_alloc_irq_vectors(pdev, 1, 1, PCI_IRQ_LEGACY /*| PCI_IRQ_MSI | PCI_IRQ_MSIX*/);
	if (err < 0) {
		dev_err(dev, "pci_alloc_irq_vectors() failed: %d\n", err);
		return err;
	}
#endif

	{
		int mac_irq = pci_irq_vector(pdev, 0);
		int clock_m_irq = pci_irq_vector(pdev, 0);
		struct device * pci_bridge = pci_data->pci_bridge ? &pci_data->pci_bridge->dev : NULL;

		struct oregano_clock_platform_data clock_platform_data = {
			.si_time_pll = pci_data->si_time_pll,
		};

		struct oregano_mac_platform_data  mac_platform_data = {
			.pci_bridge = pci_bridge,
		};

		const struct resource syn1588_pci_clock_resources[] = {
			DEFINE_PCI_RES_MEM_NAMED(PCI_OFFSET_CLOCK_M, PCI_SIZE_CLOCK_M, "main"),
			DEFINE_RES_IRQ_NAMED(clock_m_irq, "main")
		};

		const struct resource syn1588_pci_updater_resources[] = {
			DEFINE_PCI_RES_MEM_NAMED(PCI_OFFSET_FPGA_UPDATE, PCI_SIZE_FPGA_UPDATE, "main")
		};

		struct oregano_syncd_platform_data  syncd_platform_data = {
			.syn1588_dev = &pci_data->syn1588_dev,
		};

		const struct resource syn1588_pci_syncd_resources[] = {
			DEFINE_PCI_RES_MEM_NAMED(PCI_OFFSET_SPI_PLL, PCI_SIZE_SPI_PLL, "legacy-spi"),
			DEFINE_PCI_RES_MEM_NAMED(PCI_OFFSET_I2C_VIDEO_DAC, PCI_SIZE_I2C_VIDEO_DAC, "legacy-i2c-video"),
			DEFINE_PCI_RES_MEM_NAMED(PCI_OFFSET_EEPROM, PCI_SIZE_EEPROM, "legacy-eeprom"),
		};
		int i =0;

		pci_data->syn1588_dev.updater = platform_device_register_resndata(
			dev, "oregano-updater-altera", PLATFORM_DEVID_AUTO, syn1588_pci_updater_resources, ARRAY_SIZE(syn1588_pci_updater_resources),
			NULL, 0
		);
		if(IS_ERR_OR_NULL(pci_data->syn1588_dev.updater)) {
			return PTR_ERR(pci_data->syn1588_dev.updater);
		}

		pci_data->syn1588_dev.clock_m_dev = platform_device_register_resndata(
			dev, "oregano-clock", PLATFORM_DEVID_AUTO, syn1588_pci_clock_resources, ARRAY_SIZE(syn1588_pci_clock_resources),
			&clock_platform_data, sizeof(clock_platform_data)
		);
		if(IS_ERR_OR_NULL(pci_data->syn1588_dev.clock_m_dev)) {
			return PTR_ERR(pci_data->syn1588_dev.clock_m_dev);
		}

		mac_platform_data.clock_m = &pci_data->syn1588_dev.clock_m_dev->dev;
		for(i=0; i < syn1588_pci_data->num_macs && pdev->resource[i].flags != 0; i++){
			const struct resource syn1588_pci_mac_resources[] = {
				DEFINE_RES_MEM_NAMED((pdev->resource[i].start + PCI_OFFSET_MAC_MAIN), PCI_SIZE_MAC, "main"),
				DEFINE_RES_IRQ_NAMED(mac_irq, "legacy"),
				/*
				DEFINE_RES_IRQ_NAMED(mac_irq, "tx"),
				DEFINE_RES_IRQ_NAMED(mac_irq, "tx-error"),
				DEFINE_RES_IRQ_NAMED(mac_irq, "tx-ctrl"),
				DEFINE_RES_IRQ_NAMED(mac_irq, "rx"),
				DEFINE_RES_IRQ_NAMED(mac_irq, "rx-error"),
				DEFINE_RES_IRQ_NAMED(mac_irq, "rx-ctrl"),
				DEFINE_RES_IRQ_NAMED(mac_irq, "main"),
				*/
				//DEFINE_RES_IRQ_NAMED(pci_alloc_irq_vectors(pdev, SYN1588NIC_IRQ_MSI_MAC), "main")
			};
			struct platform_device *mac_dev = platform_device_register_resndata(
				dev, "oregano-mac", PLATFORM_DEVID_AUTO, syn1588_pci_mac_resources, ARRAY_SIZE(syn1588_pci_mac_resources),
				&mac_platform_data, sizeof(mac_platform_data)
			);
			if(IS_ERR_OR_NULL(mac_dev)) {
				return PTR_ERR(mac_dev);
			}
			pci_data->syn1588_dev.mac_devs[i] = mac_dev;
		}

		pci_data->syn1588_dev.sync_d = platform_device_register_resndata(
			dev, "oregano-syncd", PLATFORM_DEVID_AUTO,
			syn1588_pci_syncd_resources, id->vendor == PCI_VENDOR_ID_ALTERA ? ARRAY_SIZE(syn1588_pci_syncd_resources) : 0,
			&syncd_platform_data, sizeof(syncd_platform_data)
		);
		if(IS_ERR_OR_NULL(pci_data->syn1588_dev.sync_d)) {
			return PTR_ERR(pci_data->syn1588_dev.sync_d);
		}
	}
	dev_info(dev, "finish init syn1588_pci\n");

	return 0;
}


static void sy1588_pci_remove(struct pci_dev *pdev) {

	struct syn1588_pci_nic_device *pci_data = dev_get_drvdata(&pdev->dev);
	int i;

	if(!IS_ERR_OR_NULL(pci_data->syn1588_dev.sync_d)) {
		platform_device_unregister(pci_data->syn1588_dev.sync_d);
		pci_data->syn1588_dev.sync_d = NULL;
	}

	if(!IS_ERR_OR_NULL(pci_data->syn1588_dev.updater)) {
		platform_device_unregister(pci_data->syn1588_dev.updater);
		pci_data->syn1588_dev.updater = NULL;
	}

	for(i=0; i< MAX_AMOUNT_MACS; i++) {
		struct platform_device *mac_dev = pci_data->syn1588_dev.mac_devs[i];
		if(!IS_ERR_OR_NULL(mac_dev)) {
			platform_device_unregister(mac_dev);
			pci_data->syn1588_dev.mac_devs[i] = NULL;
		}
	}

	if(!IS_ERR_OR_NULL(pci_data->syn1588_dev.clock_m_dev)) {
		platform_device_unregister(pci_data->syn1588_dev.clock_m_dev);
		pci_data->syn1588_dev.clock_m_dev = NULL;
	}

	if(!IS_ERR_OR_NULL(pci_data->pll_spi)) {
		platform_device_unregister(pci_data->pll_spi);
		pci_data->pll_spi = NULL;
	}

	if(!IS_ERR_OR_NULL(pci_data->pci_bridge)) {
		platform_device_unregister(pci_data->pci_bridge);
		pci_data->pci_bridge = NULL;
	}

	pci_clear_master(pdev);
	if(pci_is_enabled(pdev)) {
		pci_disable_device(pdev);
	}
	pci_set_drvdata(pdev, NULL);

	pci_free_irq_vectors(pdev);
}



static pci_ers_result_t syn1588_pci_error_detected(struct pci_dev *pdev,
						   pci_channel_state_t error)
{
	struct device *dev = &pdev->dev;
	dev_alert(dev, "PCI error detected!\n");

	return PCI_ERS_RESULT_NONE;
}

static struct pci_error_handlers syn1588_pci_err_handlers = {
	.error_detected = syn1588_pci_error_detected,
};


static const struct syn1588_pci_data pci_nic_rev2 = {
	pci_rev_max_supported: 3,
	pci_rev_min_supported: 3,
	num_macs: 1,
};
static const struct syn1588_pci_data pci_nic_rev2_3 = {
	pci_rev_max_supported: 2,
	pci_rev_min_supported: 1,
	num_macs: 1,
};
static const struct syn1588_pci_data pci_dualnic_rev1_1 = {
	pci_rev_max_supported: 0,
	pci_rev_min_supported: 0,
	num_macs: 2,
};

/* This table should be in sync with the one in drivers/pci/pci-mid.c */
static const struct pci_device_id syn1588_pci_ids[] = {
	{ PCI_DEVICE_DATA(ALTERA, NIC_OLD_DEVICE_ID, &pci_nic_rev2) },
	//{ PCI_DEVICE_DATA(MEINBERG, PCI_DEVICE_ID_OREGAN_TEST_DEVICE, pci_nic_rev2_3) },
	{ PCI_DEVICE_DATA(MEINBERG, OREGANO_REV2_3_DEVICE, &pci_nic_rev2_3) },
	{ PCI_DEVICE_DATA(MEINBERG, OREGANO_DUAL_NIC_REV1_1_DEVICE, &pci_dualnic_rev1_1) },
	{}
};
MODULE_DEVICE_TABLE(pci, syn1588_pci_ids);

static struct pci_driver syn1588_pci_driver = {
	.name		= "syn1588_pci",
	.probe		= syn1588_pci_probe,
	.id_table	= syn1588_pci_ids,
	.remove		= sy1588_pci_remove,
	//.suspend = pci_dev_suspend,
	//.resume = pci_dev_resume,
	.err_handler = &syn1588_pci_err_handlers,
};



module_pci_driver(syn1588_pci_driver);

MODULE_AUTHOR("Norbert Summer");
MODULE_DESCRIPTION("Syn1588 PCI part of the oregano NIC");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);
