/*
 * Copyright (C) oregano systems 2022
 * Norbert Summer <summer@oregano.at>
 */

#ifndef SYN1588_SYS_H
#define SYN1588_SYS_H

#ifdef __cplusplus
extern "C" {
#endif

#include <linux/ioctl.h>
#include <linux/socket.h>
#include <linux/types.h>

#define SYN1588_IOC_MAGIC ((unsigned int)'s')

#define MAX_PATTERN_DATA 256


/**
 * the syn1588_timepoint type represents an absolut point in time
 * the format is time in ns since an epoch
 * the PTP epoch is normaly the 1970-01-01 00:00 tai time
 * otherwise its the time since the start of the system
*/
typedef struct {
	__u64 sec;
	__u32 nsec;
} syn1588_ioctl_timepoint;

enum SYN1588_PKTGEN_MODE {
	SYN1588_PKTGEN_MODE_L2,
	SYN1588_PKTGEN_MODE_L2_VLAN,
	SYN1588_PKTGEN_MODE_IPV4,
	SYN1588_PKTGEN_MODE_IPV4_VLAN,
	SYN1588_PKTGEN_MODE_IPV6,
	SYN1588_PKTGEN_MODE_IPV6_VLAN,
	SYN1588_PKTGEN_MODE_NTP
};

struct syn1588_ioctl_buffer {
	__u8 *data;
	__u32 size;
};

struct syn1588_ioctl_buffer_const {
	const __u8 *data;
	__u32 size;
};

enum syn1588_ioctl_update_status {
	SYN1588_IOCTL_UPDATE_STATUS_IDLE = 0,
	SYN1588_IOCTL_UPDATE_STATUS_STARTING,
	SYN1588_IOCTL_UPDATE_STATUS_PROGRESS,
	SYN1588_IOCTL_UPDATE_STATUS_PROGRESS_VERIFY,
	SYN1588_IOCTL_UPDATE_STATUS_SUCCESS,
	SYN1588_IOCTL_UPDATE_STATUS_FAILED,
};
struct syn1588_ioctl_update_status_reponse {
	enum syn1588_ioctl_update_status status;
	__u8 progress;
};



/* --------- General --------- */

#define SYN1588_IOC_RD_BUILD_ID _IOR(SYN1588_IOC_MAGIC, 10U, __u32)


/* ------ PCI ------ */

//#define SYN1588_PCI_IOC_RD_VERSION _IOR(SYN1588_IOC_MAGIC, 20U, __u8[64])
#define SYN1588_PCI_IOC_RD_MODEL _IOR(SYN1588_IOC_MAGIC, 21U, __u8[64])
//  ID 23 and 24  got deleted try not to implement does
#define SYN1588_IOC_RD_PRODUCT_ID _IOR(SYN1588_IOC_MAGIC, 25U, struct syn1588_ioctl_buffer)
#define SYN1588_PCI_IOC_WR_FLASH_AND_VERIFY _IOW(SYN1588_IOC_MAGIC, 26U, struct syn1588_ioctl_buffer_const)
#define SYN1588_IOC_RD_SERIAL_NR _IOR(SYN1588_IOC_MAGIC, 27U, struct syn1588_ioctl_buffer)
#define SYN1588_PCI_IOC_RD_UPDATE_STATUS _IOR(SYN1588_IOC_MAGIC, 28U, struct syn1588_ioctl_update_status_reponse)

// PCI - PLL
// TODO

/* ------ clock_m ------ */

#define SYN1588_CLOCK_IOC_RD_PHC_ID _IOR(SYN1588_IOC_MAGIC, 30U, __u32) // get ptp_clock_id
#define SYN1588_CLOCK_IOC_RD_VERSION _IOR(SYN1588_IOC_MAGIC, 31U, __u8[64])
#define SYN1588_CLOCK_IOC_RD_CLOCK_ID _IOR(SYN1588_IOC_MAGIC, 32U, __u8[8])

struct syn1588_ioctl_duration {
	__u64 numerator;
	__u64 denominator;
};

enum syn1588_ioctl_period_id {
	SYN1588_IOCTL_PERIOD0, // precision if required and available
	SYN1588_IOCTL_PERIOD1, // special precision if required and available
	SYN1588_IOCTL_PERIOD_DUTY0, // locks period 0 / 1 to create a period with custom duty
	SYN1588_IOCTL_PERIOD2, // light period
	SYN1588_IOCTL_PERIOD3, // light period
	__SYN1588_IOCTL_PERIOD_NBITS
};

enum syn1588_ioctl_period_start_condition {
	SYN1588_IOCTL_PERIOD_START_ENABLE,
	SYN1588_IOCTL_PERIOD_START_1PPS,
	SYN1588_IOCTL_PERIOD_START_TRIGGER0,
	SYN1588_IOCTL_PERIOD_START_TRIGGER1,
	__SYN1588_IOCTL_PERIOD_START_NBITS
};

struct syn1588_ioctl_period_config {
	__u32 id; // period id, one of {0, 1, 2, 3}
	struct syn1588_ioctl_duration high_time;
	struct syn1588_ioctl_duration low_time;
	__u8 is_rising_edge; // bool
	__u8 start_condition; // enum syn1588_ioctl_period_start_condition
};

struct syn1588_ioctl_period_set_enable {
	__u32 id;
	__u8 is_enabled; // bool
};

struct syn1588_ioctl_period_get_duration {
	__u32 id;
	struct syn1588_ioctl_duration high_time;
	struct syn1588_ioctl_duration low_time;
};

enum syn1588_ioctl_event_id {
	SYN1588_IOCTL_EVENT0, // better fast trigger capable
	SYN1588_IOCTL_EVENT1,
	__SYN1588_IOCTL_EVENT_NBITS
};
struct syn1588_ioctl_event_read_action {
	__u32 id; // enum syn1588_ioctl_event_id
	__u32 bufsize; // get filled by the amount of gotten timestamps
	syn1588_ioctl_timepoint *buf; // array of times in ns
};

enum syn1588_ioctl_trigger_id {
	SYN1588_IOCTL_TRIGGER0,
	SYN1588_IOCTL_TRIGGER1,
	__SYN1588_IOCTL_TRIGGER_NBITS
};

enum syn1588_ioctl_trigger_action {
	SYN1588_IOCTL_TRIGGER_ACTION_LOW,
	SYN1588_IOCTL_TRIGGER_ACTION_HIGH,
	SYN1588_IOCTL_TRIGGER_ACTION_TOGGLE
};

struct syn1588_ioctl_trigger_entry {
	syn1588_ioctl_timepoint ts;
	enum syn1588_ioctl_trigger_action action;
};
struct syn1588_ioctl_trigger_add_actions {
	__u32 id; // enum syn1588_ioctl_event_id
	__u32 bufsize; // get filled by the amount of written timestamps
	struct syn1588_ioctl_trigger_entry *buf; // array of time points
};

enum syn1588_ioctl_io_port_id {
	SYN1588_IOCTL_IO_PORT_X4,
	SYN1588_IOCTL_IO_PORT_X5,
	SYN1588_IOCTL_IO_PORT_X6,
	SYN1588_IOCTL_IO_PORT_X7,
	__SYN1588_IOCTL_IO_PORT_NBITS
};


enum syn1588_ioctl_source_type {
	SYN1588_IOCTL_SOURCE_PERIOD,
	SYN1588_IOCTL_SOURCE_TRIGGER,
	SYN1588_IOCTL_SOURCE_IO_PORT,
	SYN1588_IOCTL_SOURCE_IRIG_B,
	SYN1588_IOCTL_SOURCE_1PPS,
	SYN1588_IOCTL_SOURCE_PLL_FILTERED,
	__SYN1588_IOCTL_SOURCE_NBITS
};
#define SYN1588_IOCTL_MAX_SOURCE_ID MAX(__SYN1588_IOCTL_PERIOD_NBITS, MAX(__SYN1588_IOCTL_TRIGGER_NBITS, __SYN1588_IOCTL_IO_PORT_NBITS))

enum syn1588_ioctl_sink_type {
	SYN1588_IOCTL_SINK_EVENT,
	SYN1588_IOCTL_SINK_IO_PORT,
	SYN1588_IOCTL_SINK_IRIG_B,
	__SYN1588_IOCTL_SINK_NBITS
};
#define SYN1588_IOCTL_MAX_SINK_ID MAX(__SYN1588_IOCTL_EVENT_NBITS, __SYN1588_IOCTL_IO_PORT_NBITS)


struct syn1588_ioctl_source {
	enum syn1588_ioctl_source_type type;
	int id;
};
struct syn1588_ioctl_sink {
	enum syn1588_ioctl_sink_type type;
	int id;
};

#define SYN1588_CLOCK_IOC_WR_PERIOD_ALLOC _IOW(SYN1588_IOC_MAGIC, 33U, struct syn1588_ioctl_period_config)
#define SYN1588_CLOCK_IOC_WR_PERIOD_FREE _IOW(SYN1588_IOC_MAGIC, 34U, __u32 )
#define SYN1588_CLOCK_IOC_WR_PERIOD_GET_DURATION _IOW(SYN1588_IOC_MAGIC, 35U, struct syn1588_ioctl_period_get_duration )
// Enables / Disables a period function
#define SYN1588_CLOCK_IOC_WR_PERIOD_SET_ENABLE _IOW(SYN1588_IOC_MAGIC, 36U, struct syn1588_ioctl_period_set_enable )

#define SYN1588_CLOCK_IOC_WR_EVENT_ALLOC _IOW(SYN1588_IOC_MAGIC, 37U, __u32)
#define SYN1588_CLOCK_IOC_WR_EVENT_FREE _IOW(SYN1588_IOC_MAGIC, 38U, __u32 )
#define SYN1588_CLOCK_IOC_WR_EVENT_READ_TS _IOW(SYN1588_IOC_MAGIC, 39U, struct syn1588_ioctl_event_read_action )
#define SYN1588_CLOCK_IOC_WR_EVENT_CLEAR _IOW(SYN1588_IOC_MAGIC, 47U, __u32 )

#define SYN1588_CLOCK_IOC_WR_TRIGGER_ALLOC _IOW(SYN1588_IOC_MAGIC, 40U, __u32)
#define SYN1588_CLOCK_IOC_WR_TRIGGER_FREE _IOW(SYN1588_IOC_MAGIC, 41U, __u32 )
#define SYN1588_CLOCK_IOC_WR_TRIGGER_ADD_ACTIONS _IOW(SYN1588_IOC_MAGIC, 42U, struct syn1588_ioctl_trigger_add_actions )

#define SYN1588_CLOCK_IOC_WR_IO_PORT_ALLOC _IOW(SYN1588_IOC_MAGIC, 43U, __u32)
#define SYN1588_CLOCK_IOC_WR_IO_PORT_FREE _IOW(SYN1588_IOC_MAGIC, 44U, __u32 )


struct syn1588_ioctl_connection {
	struct syn1588_ioctl_source source;
	struct syn1588_ioctl_sink sink;
};

#define SYN1588_CLOCK_IOC_WR_IO_MATRIX_CONNECT _IOW(SYN1588_IOC_MAGIC, 45U, struct syn1588_ioctl_connection )
#define SYN1588_CLOCK_IOC_WR_IO_MATRIX_DISCONNECT _IOW(SYN1588_IOC_MAGIC, 46U, struct syn1588_ioctl_sink )
/*
TODO
#define SYN1588_CLOCK_IOC_WR_1PPS_ALLOC _IOW(SYN1588_IOC_MAGIC, 42U, __u32)
#define SYN1588_CLOCK_IOC_WR_1PPS_FREE _IOW(SYN1588_IOC_MAGIC, 43U, __u32 )

#define SYN1588_CLOCK_IOC_WR_IRIG_B_ALLOC _IOW(SYN1588_IOC_MAGIC, 42U, __u32)
#define SYN1588_CLOCK_IOC_WR_IRIG_B_FREE _IOW(SYN1588_IOC_MAGIC, 43U, __u32 )

#define SYN1588_CLOCK_IOC_WR_PLL_ALLOC _IOW(SYN1588_IOC_MAGIC, 42U, __u32)
#define SYN1588_CLOCK_IOC_WR_PLL_FREE _IOW(SYN1588_IOC_MAGIC, 43U, __u32 )
*/

// look if it is possible to implement this in the SO_TIMESTAMPING api
#define SYN1588_CLOCK_IOC_RD_TIMESTAMPER_LIST \
	_IOW(SYN1588_IOC_MAGIC, 50U, struct syn1588_timestamp_config)
#define SYN1588_CLOCK_IOC_RD_TIMESTAMPER_CONFIG \
	_IOW(SYN1588_IOC_MAGIC, 51U, struct syn1588_timestamp_config)
#define SYN1588_CLOCK_IOC_WR_TIMESTAMPER_CONFIG \
	_IOW(SYN1588_IOC_MAGIC, 52U, struct syn1588_timestamp_config)
#define SYN1588_CLOCK_IOC_RD_TIMESTAMP \
	_IORW(SYN1588_IOC_MAGIC, 53U, struct timestamp)

//iomatrix
//TODO

/* ------ mac ------ */

#define SYN1588_NIC_IOC_RD_VERSION _IOR(SYN1588_IOC_MAGIC, 81U, __u32)

// look if it is possible to implement this in the SO_TIMESTAMPING api
#define SYN1588_NIC_IOC_WR_TIMESTAMP_CONFIG									\
	_IOW(SYN1588_IOC_MAGIC, 82U, struct syn1588_timestamp_config)
#define SYN1588_NIC_IOC_GET_TIMESTAMP										  \
	_IORW(SYN1588_IOC_MAGIC, 82U,										  \
		  struct timestamp) // NO SO_TIMESTAMPING Systems

#ifdef __cplusplus
}
#endif

#endif // end SYN1588_SYS_H
